from .arguments import CUSTOM_ARGUMENTS_TYPES, CUSTOM_TYPES_MAPPING
from .decorators import dataclass_argument
from .parser import ArgumentsParser


__all__ = [
    'ArgumentsParser',
    'CUSTOM_ARGUMENTS_TYPES',
    'CUSTOM_TYPES_MAPPING',
    'dataclass_argument',
]



import os
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Optional, Union

import toml
import yaml

from script_args_parser.arguments import Argument, argument_factory


class ArgumentsParser:
    """
    Parses arguments according to given toml definition and cli parameters.
    Values for arguments are stored in arguments_values dictionary.

    :param arguments_definitions: toml string containing arguments definition
    :param cli_params: list of cli parameters, if not given sys.arg[1:] is used
    """
    def __init__(
        self, arguments: list[Argument], cli_params: Optional[list[str]] = None,
        user_values: Optional[dict[str, Any]] = None
    ) -> None:
        self.user_values = user_values or {}
        self.arguments = arguments
        self.arguments_values = self._read_cli_arguments(cli_params)
        self._fallback_values()
        self._parse_values()
        self._convert_values()
        self._validate_required()
        self._post_process()

    def __getattr__(self, name: str) -> Any:
        if name != 'arguments_values' and name in self.arguments_values:
            return self.arguments_values[name]
        raise AttributeError(f'No attribute named "{name}"')

    def __setattr__(self, name: str, value: Any) -> None:
        if name != 'arguments_values' and name in getattr(self, 'arguments_values', {}):
            self.arguments_values[name] = value
        else:
            super().__setattr__(name, value)

    @classmethod
    def from_files(
        cls, arguments_file: Union[str, Path], cli_params: Optional[list[str]] = None,
        yaml_config: Optional[Union[str, Path]] = None
    ) -> 'ArgumentsParser':
        if isinstance(arguments_file, str):
            arguments_file = Path(arguments_file)
        if isinstance(yaml_config, str):
            yaml_config = Path(yaml_config)
        arguments = cls._parse_toml_definitions(arguments_file.read_text())
        if yaml_config is None:
            user_values = None
        else:
            user_values = yaml.load(yaml_config.read_text(), Loader=yaml.SafeLoader)
        return cls(arguments, cli_params, user_values)

    @staticmethod
    def _parse_toml_definitions(toml_string: str) -> list[Argument]:
        parsed_toml = toml.loads(toml_string)
        return [argument_factory(arg_name, arg_def) for arg_name, arg_def in parsed_toml.items()]

    def _read_cli_arguments(self, cli_params: list[str] = None) -> dict[str, Any]:
        cli_parser = ArgumentParser()
        for argument in self.arguments:
            args, kwargs = argument.argparse_options
            cli_parser.add_argument(*args, **kwargs)
        return vars(cli_parser.parse_args(cli_params))

    def _fallback_values(self) -> None:
        for argument in self.arguments:
            if self.arguments_values[argument.name] is None:
                self.arguments_values[argument.name] = self.user_values.get(argument.name)
            if self.arguments_values[argument.name] is None and argument.env_var is not None:
                self.arguments_values[argument.name] = os.getenv(argument.env_var)
            if self.arguments_values[argument.name] is None and argument.default_value is not None:
                self.arguments_values[argument.name] = argument.default_value

    def _parse_values(self) -> None:
        for argument in self.arguments:
            if (argument_value := self.arguments_values[argument.name]) is not None:
                self.arguments_values[argument.name] = argument.parse_value(argument_value)

    def _convert_values(self) -> None:
        for argument in self.arguments:
            if (argument_value := self.arguments_values[argument.name]) is not None:
                self.arguments_values[argument.name] = argument.convert_value(argument_value)

    def _validate_required(self) -> None:
        for arg in self.arguments:
            if arg.required and self.arguments_values[arg.name] is None:
                error_msg = f'No value supplied for argument "{arg.name}". You can set it in config file'
                if arg.cli_arg is not None:
                    error_msg += f' or by using cli option: "{arg.cli_arg}"'
                if arg.env_var is not None:
                    error_msg += f' or by setting env variable: "{arg.env_var}"'
                error_msg += '.'
                raise RuntimeError(error_msg)

    def _post_process(self) -> None:
        for argument in self.arguments:
            self.arguments_values[argument.name] = argument.post_process(
                self.arguments_values[argument.name],
                self.arguments_values,
            )



from typing import Any, Type, Union

from script_args_parser.arguments import CUSTOM_TYPES_MAPPING


def dataclass_argument(decorated: Type) -> Type:
    def argument_factory(definition: Union[dict, list]) -> Any:
        if isinstance(definition, dict):
            return decorated(**definition)
        if isinstance(definition, list):
            return decorated(*definition)
    decorated.__argument_factory = argument_factory
    CUSTOM_TYPES_MAPPING[decorated.__name__] = decorated.__argument_factory
    return decorated



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['script_args_parser']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0', 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'script-args-parser',
    'version': '0.4.0',
    'description': 'Script arguments parsing library.',
    'long_description': '# Script Arguments Parser\n\nThis library is meant to provide an easy way to consume arguments for scripts in more complex scenarios without writing too much code.\n\n[![license](https://img.shields.io/github/license/KRunchPL/script-args-parser.svg)](https://github.com/KRunchPL/script-args-parser/blob/master/LICENSE)\n[![latest release](https://img.shields.io/github/release/KRunchPL/script-args-parser.svg)](https://github.com/KRunchPL/script-args-parser/releases/latest) [![latest release date](https://img.shields.io/github/release-date/KRunchPL/script-args-parser.svg)](https://github.com/KRunchPL/script-args-parser/releases)\n\n[![PyPI version](https://img.shields.io/pypi/v/script-args-parser)](https://pypi.org/project/script-args-parser/) [![Python](https://img.shields.io/pypi/pyversions/script-args-parser)](https://pypi.org/project/script-args-parser/)\n\n\n## Why something more?\n\nIn Python there are a lot of ways to consume cli parameters, starting from built-in parsers finishing at libraries like docopt. But unfortunately during my adventure I encountered a few problems that were not solvable just by using one of them. Few of those problems:\n\n* get values from multiple sources: cli, config file, environment variable, default;\n* convert given variable according to argument definition;\n* all argument information (cli option, fallback env var, conversion type, default value etc.) defined in one place;\n* definitions written outside the code, so the script is kept clean and simple;\n* more complex conversion types build in.\n\n## Main features\n\n* Parameters defined in both human- and computer-readable format outside of the code, in one place\n* Argument values converted to given format (predefined or custom)\n* Config file fallback\n* Environmental variable fallback\n* Default values\n* Human readable errors\n\n## Usage\n\nOne of the goals of this library was to minimize amount of the code. Therefore whole usage looks like this:\n\n```python\nfrom script_args_parser import ArgumentsParser\n\nargs = ArgumentsParser.from_files(\'example-parameters.toml\', yaml_config=\'example-config.yaml\')\nprint(args.name)\nprint(args.age)\n```\n\nAbove script will read arguments definition from `example-parameters.toml` and try to read their values in following order:\n\n1. from cli options,\n2. from config file (`example-config.yaml` in example),\n3. from environment variables,\n4. default values.\n\nIf any argument does not have value defined it will be None, unless it is required, so it will raise an exception.\n\nWhen all values are established, parser will convert them to specified type.\n\n### Arguments definition\n\nThe list of script arguments is provided in toml file. Example argument can look like this:\n```toml\n[name]\ndescription = "Some fancy description"  # required\ntype = "str"   # required\ncli_arg = "--cli-opt"  # required\nenv_var = "ENV_VAR_NAME"\nrequired = false\ndefault_value = "I got you"\n```\n\n#### description **(mandatory)**\n\nHuman readable description of an argument.\n\n#### type **(mandatory)**\n\nParser will use this field to convert value of the argument from string to one that is specified.\n\nSome more complex types are also changing the way cli options are parsed.\n\nFor detailed description of possible values and their meaning, see [Types section](#types).\n\n#### cli_arg **(mandatory)**\n\nName of the cli option throught which value can be set.\n\n#### env_var\n\nName of environment variable that will be used to read value if not specified by CLI or config file.\n\nFor the format used by more complex types see [Types section](#types).\n\n#### required\n\nBy default False. If set to true, the parser will raise an error if value will not be found anywhere.\n\nCan be specified as boolean value (true, false) or by string (\'true\', \'false\', \'yes\', \'no\', \'1\', \'0\').\n\n#### default_value\n\nValue that will be used if not specified by CLI, config file or environment variable.\n\nFor the format used by more complex types see [Types section](#types).\n\n### Types\n\nThis is the list of built-in types supported.\n\n#### String\n\nType field value: `str`\n\nNo special operations are performed.\n\n#### Integer\n\nType field value: `int`\n\nValue will be parsed to integer, if not possible, exception will be raised.\n\n##### Post operations\n\nAdditional parameter `post_operations` can be used. It stores the expression that will be evaluated after the value is read. The result of evaluation will be used as a value. The `{value}` token in expression will be substituted with the value provided by the user.\n\nFor example when a program requires value in seconds, but the user will always want to specify minutes the `post_operations` can be: `"{value} * 60"`.\n\n#### Boolean\n\nType field value: `bool`\n\nSome strings has been defined to be matched to specific values (case insensitive):\n\n* True can be specified as: true, yes, 1;\n* False can be specified as: false, no, 0;\n\nAll other values will be converted to bool using Python rules.\n\n#### Switch\n\nType field value: `switch`\n\nBehaves in the same way as `bool` but additionaly cli option can be passed without an argument and will be considered True.\n\n#### Path\n\nType field value: `path`\n\nWill be converted into `pathlib.Path` object. Worth noticing is that empty string will be equivalent of current directory.\n\n##### Parent path\n\nAdditional parameter `parent_path` can be used. It shall contain a name of another path argument. Current path will be prepended with the value of `parent_path`.\n\nWith given toml file the default value of `picture_name` will be `\'images/beautiful.jpg\'`.\n\n```toml\n[pictures_folder]\ndescription = "Path to folder with pictures"\ntype = "path"\ncli_arg = "--pictures-folder"\ndefault = "./images"\n\n[picture_name]\ndescription = "Name of a picture file"\ntype = "path"\ncli_arg = "--picture-name"\nparent_path = "pictures_folder"\ndefault = "beautiful.jpg"\n```\n\nIt is possible to make a hierarchy of paths, but keep in mind that the arguments are evaluated in order that they are defined in toml file, so with below toml file the `picture_name` will have value `\'the_best_user\\beautiful.jpg\'` even though `user_folder` will be `pictures_folder\\user_folder`.\n\n```toml\n[pictures_folder]\ndescription = "Path to folder with pictures"\ntype = "path"\ncli_arg = "--pictures-folder"\ndefault = "./images"\n\n[picture_name]\ndescription = "Name of a picture file"\ntype = "path"\ncli_arg = "--picture-name"\nparent_path = "user_folder"\ndefault = "beautiful.jpg"\n\n[user_folder]\ndescription = "Name of a user folder file"\ntype = "path"\ncli_arg = "--user-folder"\nparent_path = "pictures_folder"\ndefault = "the_best_user"\n```\n\n#### List\n\nType field value: `list[<simple type>]`\n\nWill produce a list of elements with given simple types (any that was described above).\n\nWhen this type is specified, multiple cli options should be used to pass list elements:\n```\nscript.py --child-name John --child-name David\n```\n\nIn default value or environment variable use semicolon to split values:\n```\ndefault_value = "John; David; \'Some;Very;Strange;Name\'"\n```\n\n#### Tuple\n\nType field value: `tuple[<simple type>, <optional simple type>, ...]`\n\nExample type field value: `tuple[str]`, `tuple[int, str, bool]`.\n\nWill produce a list with given amount of values of simple types elements.\n\nWhen this type is specified, cli options should be used once but with multiple values. For  `tuple[str, str, str]`\n```\nscript.py --all-my-names John Maria "De\'naban"\n```\n\nIn default value or environment variable separate values with space:\n```\ndefault_value = "John Maria "De\'naban"\n```\n\n#### List of tuples\n\nType field value: `list[tuple[<simple type>, <optional simple type>, ...]]`\n\nCombining list and tuple types. Will produce a list of lists.\n\nFor cli use:\n```\nscript.py --child John 16 --child David 18 --child Maria 21\n```\n\nFor default values and enviroment variables use:\n```\ndefault_value = "John 16; David 18; Maria 21"\n```\n\nAbove examples for `list[tuple[str, int]]` will produce:\n\n```python\n[[\'John\', 16], [\'David\', 18], [\'Maria\', 21]]\n```\n\n#### Dataclass argument\n\nType field value: `<name of the dataclass>`\n\nParses a list or a dict to user defined dataclass.\n\nIn order for the dataclass to be used it has to be decorated with `script_args_parser.decorators.dataclass_argument`.\n\nWith given Python code and toml definition:\n\n```python\n@dataclass_argument\n@dataclass\nclass MyDataClass:\n    value_1: str\n    value_2: str\n```\n\n```toml\n[two_values]\ndescription = "Some two string values"\ntype = "MyDataClass"\ncli_arg = "--two-values"\n```\n\nThe following yaml input files can be used\n\n```yaml\ntwo_values:\n  - first_value\n  - second value\n```\n\nor\n\n```yaml\ntwo_values:\n  value_1: first_value\n  value_2: second value\n```\n\nThe type can also be used as a list type argument like:\n\n```toml\n[two_values]\ndescription = "Some two string values"\ntype = "list[MyDataClass]"\ncli_arg = "--two-values"\n```\n\nNOTE: Currently cli or env values are not supported for this type.\n\n## Planned work\n\nWork that still need to be done prior to v1.0\n\n- [x] Default and envs for list\n- [x] Default and envs for tuple\n- [x] Default and envs for list of tuples\n- [x] Add more list of tuples tests\n- [x] Add path type (with tests)\n- [x] Create from path\n- [x] Support config file\n- [x] Document possible types\n- [ ] Add support for env and cli values for dataclass type\n- [ ] Write some complex test cases\n- [ ] Allow non-cli arguments\n- [ ] Add logging\n- [ ] Allow custom argument types\n- [ ] Generate usage\n- [ ] Error handling\n- [ ] TOML file validation\n- [ ] CI/CD\n\n## Contributing\n\nRight now I would like to finish what I planned by myself and release version 1.0. If you have any suggestions or you have found bugs, feel free to submit an issue and I will take a look at it as soon as possible.\n\n## Development\n\nDevelopment documentation can be found [here](README-DEV.md)\n',
    'author': 'KRunchPL',
    'author_email': 'krunchfrompoland@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/KRunchPL/script-args-parser',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import re
import shlex
from copy import deepcopy
from dataclasses import asdict, dataclass
from itertools import chain
from pathlib import Path
from typing import Any, Callable, Optional, Union


def _split_string_by_semicolon(argument_value: str) -> list[Any]:
    if argument_value == '':
        return ['']
    argument_value = ' ' + argument_value + ' '
    while argument_value.find(';;') != -1:
        argument_value = argument_value.replace(';;', '; ;', 1)
    parser = shlex.shlex(argument_value)
    parser.whitespace_split = True
    parser.whitespace = ';'
    return list(parser)


def _str_to_bool(value: str) -> bool:
    """
    Parses string into bool. It tries to match some predefined values.
    If none is matches, python bool(value) is used.

    :param value: string to be parsed into bool
    :return: bool value of a given string
    """
    if isinstance(value, str):
        if value.lower() in ['0', 'false', 'no']:
            return False
        if value.lower() in ['1', 'true', 'yes']:
            return True
    return bool(value)


_BASIC_TYPES_MAPPING: dict[str, Callable] = {
    'str': str,
    'int': int,
    'bool': _str_to_bool,
    'path': Path,
}  #: Built-in map: string value -> types to actual converter


CUSTOM_TYPES_MAPPING: dict[str, Callable] = {
}  #: Maps string values of types to actual converters


@dataclass
class Argument:
    """
    Stores information about single script arguments
    """
    name: str  #: custom name that the value will be stored at
    type: str  #: one of supported types (see README.md)
    description: str  #: user friendly description
    cli_arg: str  #: name of the cli option to set value
    required: bool = False  #: if set to True and the field is not set in any way, exception should be raised
    env_var: Optional[str] = None  #: name of the env var that will be used as a fallback when cli not set
    default_value: Optional[str] = None  #: default value if nothing else is set

    def __post_init__(self):
        if isinstance(self.required, str):
            self.required = _str_to_bool(self.required)

    def parse_value(self, argument_value: Any) -> Any:
        return argument_value

    def convert_value(self, argument_value: Any) -> Any:
        return self.types_mapping[self.type](argument_value)

    def post_process(self, argument_value: Any, arguments: dict[str, Any]) -> Any:
        return argument_value

    @property
    def argparse_options(self) -> dict:
        """
        :return: args and kwargs that can be used in argparse.ArgumentParser.add_argument
        """
        args = [self.cli_arg]
        kwargs = {'dest': self.name}
        if self.type == 'switch':
            kwargs['action'] = 'store'
            kwargs['nargs'] = '?'
            kwargs['const'] = True
        return (args, kwargs)

    @property
    def types_mapping(self) -> dict[str, Callable]:
        ret_val = deepcopy(_BASIC_TYPES_MAPPING)
        ret_val.update(CUSTOM_TYPES_MAPPING)
        return ret_val

    @staticmethod
    def matcher(arg_type: str) -> bool:
        lower_arg_type = arg_type.lower()
        return (
            arg_type in chain(_BASIC_TYPES_MAPPING, CUSTOM_TYPES_MAPPING)
            or
            lower_arg_type in chain(_BASIC_TYPES_MAPPING, CUSTOM_TYPES_MAPPING)
        )


@dataclass
class SwitchArgument(Argument):
    @property
    def argparse_options(self) -> dict:
        """
        :return: args and kwargs that can be used in argparse.ArgumentParser.add_argument
        """
        args = [self.cli_arg]
        kwargs = {
            'dest': self.name,
            'action': 'store',
            'nargs': '?',
            'const': True,
        }
        return (args, kwargs)

    def convert_value(self, argument_value: Any) -> bool:
        return _str_to_bool(argument_value)

    @staticmethod
    def matcher(arg_type: str) -> bool:
        return arg_type.lower() == 'switch'


@dataclass
class PathArgument(Argument):
    parent_path: Optional[str] = None  #: name of an argument holding parent path

    def convert_value(self, argument_value: Any) -> Path:
        return Path(argument_value)

    @staticmethod
    def matcher(arg_type: str) -> bool:
        return arg_type.lower() == 'path'

    def post_process(self, argument_value: Path, arguments: dict[str, Any]) -> Path:
        if self.parent_path is not None:
            if not isinstance(parent_path_value := arguments.get(self.parent_path), Path):
                raise ValueError(f'Parent path has to be a Path not {type(parent_path_value)}')
            return parent_path_value / argument_value
        else:
            return argument_value


@dataclass
class IntArgument(Argument):
    post_operations: Optional[str] = None  #: expression to calulate final value; {value} will be substituted

    def convert_value(self, argument_value: Any) -> int:
        return int(argument_value)

    @staticmethod
    def matcher(arg_type: str) -> bool:
        return arg_type.lower() == 'int'

    def post_process(self, argument_value: int, arguments: dict[str, Any]) -> int:
        if self.post_operations is not None:
            return eval(self.post_operations.format(value=argument_value))
        else:
            return argument_value


class ListArgument(Argument):
    _TYPE_REGEX = re.compile(r'list\[(.+)\]')

    def __post_init__(self):
        super().__post_init__()
        match = self._TYPE_REGEX.match(self.type)
        if match is None:
            raise ValueError(f'List type has to match regexp {self._TYPE_REGEX.pattern}. Found {self.type}.')
        self.items_type = match[1]

    def parse_value(self, argument_value: Union[str, list]) -> list[Any]:
        if isinstance(argument_value, list):
            return argument_value
        elif not isinstance(argument_value, str):
            raise TypeError(
                f'Value for list type has to be either list or string. Found {type(argument_value)}.'
            )
        ret_val = []
        for value in _split_string_by_semicolon(argument_value):
            parsed_value = shlex.split(value)
            if len(parsed_value) == 0:
                ret_val.append('')
            else:
                ret_val.append(parsed_value[0])
        return ret_val

    def convert_value(self, argument_value: list[Any]) -> list[Any]:
        return [
            self.types_mapping[self.items_type](x) for x in argument_value
        ]

    @property
    def argparse_options(self) -> dict:
        """
        :return: args and kwargs that can be used in argparse.ArgumentParser.add_argument
        """
        args = [self.cli_arg]
        kwargs = {
            'dest': self.name,
            'action': 'append',
        }
        return (args, kwargs)

    @staticmethod
    def matcher(arg_type: str) -> bool:
        return arg_type.lower().startswith('list[')


class TupleArgument(Argument):
    _TYPE_REGEX = re.compile(r'tuple\[(.+)\]')

    def __post_init__(self):
        super().__post_init__()
        match = self._TYPE_REGEX.match(self.type)
        if match is None:
            raise ValueError(f'Tuple type has to match regexp {self._TYPE_REGEX.pattern}. Found {self.type}.')
        self.items_types = [x.strip() for x in match[1].split(',')]

    def parse_value(self, argument_value: Union[str, list]) -> list[Any]:
        if isinstance(argument_value, list):
            return argument_value
        elif not isinstance(argument_value, str):
            raise TypeError(
                f'Value for tuple type has to be either list or string. Found {type(argument_value)}.'
            )
        ret_val = shlex.split(argument_value)
        if len(ret_val) == 0:
            return ['']
        expected_number = len(self.items_types)
        actual_number = len(ret_val)
        if actual_number != expected_number:
            raise RuntimeError(
                f'Tuple {self.name} expected {expected_number} values and got {actual_number}: '
                f'{argument_value}.'
            )
        return ret_val

    def convert_value(self, argument_value: list[Any]) -> list[Any]:
        converters = [self.types_mapping[x] for x in self.items_types]
        return [
            conv(value) for conv, value in zip(converters, argument_value)
        ]

    @property
    def argparse_options(self) -> dict:
        """
        :return: args and kwargs that can be used in argparse.ArgumentParser.add_argument
        """
        args = [self.cli_arg]
        kwargs = {
            'dest': self.name,
            'nargs': len(self.items_types),
        }
        return (args, kwargs)

    @staticmethod
    def matcher(arg_type: str) -> bool:
        return arg_type.lower().startswith('tuple[')


class ListOfTuplesArgument(Argument):
    _TYPE_REGEX = re.compile(r'list\[(tuple\[(.+)\])\]')

    def __post_init__(self):
        super().__post_init__()
        match = self._TYPE_REGEX.match(self.type)
        if match is None:
            raise ValueError(
                f'List of tuples type has to match regexp {self._TYPE_REGEX.pattern}. Found {self.type}.'
            )
        definition = asdict(self)
        definition['type'] = match[1]
        self.tuple_argument = TupleArgument(**definition)

    def parse_value(self, argument_value: Union[str, list]) -> list[list[Any]]:
        if isinstance(argument_value, list):
            return argument_value
        elif not isinstance(argument_value, str):
            raise TypeError(
                f'Value for list of tuples has to be either list or string. Found {type(argument_value)}.'
            )
        ret_val = []
        for value in _split_string_by_semicolon(argument_value):
            ret_val.append(self.tuple_argument.parse_value(value))
        return ret_val

    def convert_value(self, argument_value: list[list[Any]]) -> list[list[Any]]:
        return [self.tuple_argument.convert_value(item_value) for item_value in argument_value]

    @property
    def argparse_options(self) -> dict:
        """
        :return: args and kwargs that can be used in argparse.ArgumentParser.add_argument
        """
        args = [self.cli_arg]
        kwargs = {
            'dest': self.name,
            'nargs': len(self.tuple_argument.items_types),
            'action': 'append',
        }
        return (args, kwargs)

    @staticmethod
    def matcher(arg_type: str) -> bool:
        return arg_type.lower().startswith('list[tuple[')


_BUILT_IN_ARGUMENTS_TYPES = [
    ListOfTuplesArgument, ListArgument, TupleArgument, SwitchArgument, PathArgument, IntArgument, Argument
]
CUSTOM_ARGUMENTS_TYPES = []


def argument_factory(name: str, definition: dict) -> Argument:
    for argument_class in chain(CUSTOM_ARGUMENTS_TYPES, _BUILT_IN_ARGUMENTS_TYPES):
        if argument_class.matcher(definition['type']):
            return argument_class(name=name, **definition)
    raise ValueError(f'Unknown argument type: {definition["type"]}')
